﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace Kompozyt
{
    #region Abstrakcje
    public interface IPracownik
    {
        //void WyświetlInformacje(int głębokość);
        void PrzyjmijWizytę(IOdwiedzający odwiedzający, int głębokość);
    }

    public interface IOdwiedzający
    {
        void Odwiedź(IPracownik pracownik, int głębokość); //mógłbym --> zdarzenie
    }

    #endregion

    public class Pracownik : IPracownik //liść
    {
        private string imię, nazwisko, stanowisko;
        protected internal bool odwiedzony = false; //flaga rozwiązuja problem cykli

        public Pracownik(string imię, string nazwisko, string stanowisko)
        {
            this.imię = imię;
            this.nazwisko = nazwisko;
            this.stanowisko = stanowisko;
        }

        public override string ToString()
        {
            return imię + " " + nazwisko + " (" + stanowisko + ")";
        }

        /*
        public virtual void WyświetlInformacje(int głębokość)
        {
            zróbWcięcie(głębokość);
            Console.WriteLine(ToString());
            odwiedzony = true;
        }
        */
        public virtual void PrzyjmijWizytę(IOdwiedzający odwiedzający, int głębokość)
        {
            odwiedzający.Odwiedź(this, głębokość);
            odwiedzony = true;
        }

        /*
        private void zróbWcięcie(int głębokość)
        {
            for (int i = 0; i < głębokość; ++i)
                Console.Write("  ");
        }
        */

        public virtual void Resetuj()
        {
            odwiedzony = false;
        }
    }

    public class Kierownik : Pracownik, IEnumerable<IPracownik> //gałąź, kompozyt = composite
    {
        private List<IPracownik> podwładni = new List<IPracownik>();

        public Kierownik(string imię, string nazwisko, string stanowisko)
            : base(imię, nazwisko, stanowisko)
        { }

        public void DodajPodwładnego(IPracownik pracownik)
        {
            podwładni.Add(pracownik);
        }

        private static int? głębokośćNajwyższegoKierownika = null;

        /*
        public override void WyświetlInformacje(int głębokość = 0)
        {
            if (!głębokośćNajwyższegoKierownika.HasValue)
                głębokośćNajwyższegoKierownika = 0;

            base.WyświetlInformacje(głębokość);
            Console.WriteLine("Podwładni:");
            //foreach(IPracownik podwładny in podwładni)
            foreach (Pracownik podwładny in podwładni) //TODO: IPracownik
            {
                if (!podwładny.odwiedzony)
                    podwładny.WyświetlInformacje(głębokość + 1);
                else Console.WriteLine("Wykryto cykl");
            }

            if (głębokość == głębokośćNajwyższegoKierownika) Resetuj();
        }
        */

        public override void PrzyjmijWizytę(IOdwiedzający odwiedzający, int głębokość = 0)
        {
            if (!głębokośćNajwyższegoKierownika.HasValue)
                głębokośćNajwyższegoKierownika = 0;

            base.PrzyjmijWizytę(odwiedzający, głębokość);
            foreach (Pracownik podwładny in podwładni)
            {
                if (!podwładny.odwiedzony)
                    podwładny.PrzyjmijWizytę(odwiedzający, głębokość + 1);
            }

            if (głębokość == głębokośćNajwyższegoKierownika) Resetuj();
        }

        public override void Resetuj()
        {
            base.Resetuj();
            foreach (Pracownik podwładny in podwładni)
            {
                if (podwładny.odwiedzony)
                    podwładny.Resetuj();
            }
        }

        #region Wyliczanie
        public List<IPracownik> ToList()
        {
            OdwiedzającySpłaszczający odwiedzającySpłaszczający = new OdwiedzającySpłaszczający();
            PrzyjmijWizytę(odwiedzającySpłaszczający);
            return odwiedzającySpłaszczający.ListaPracowników;            
        }


        public IEnumerator<IPracownik> GetEnumerator()
        {
            return ToList().GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
        #endregion
    }

    public class OdwiedzającyWyświetlającyInformacje : IOdwiedzający
    {
        private void zróbWcięcie(int głębokość)
        {
            for (int i = 0; i < głębokość; ++i)
                Console.Write("  ");
        }

        public void Odwiedź(IPracownik pracownik, int głębokość)
        {
            zróbWcięcie(głębokość);
            Console.WriteLine(pracownik.ToString());
        }
    }

    public class OdwiedzającyLiczący : IOdwiedzający
    {
        private int licznikOdwiedzin = 0;
        public void Odwiedź(IPracownik pracownik, int głębokość)
        {
            licznikOdwiedzin++;
        }

        public void ZwróćLiczbęOdwiedzin()
        {
            Console.WriteLine("Liczba odwiedzin: " + licznikOdwiedzin);
        }
    }

    public class OdwiedzającySpłaszczający : IOdwiedzający
    {
        private List<IPracownik> pracownicy = new List<IPracownik>();

        public void Odwiedź(IPracownik pracownik, int głębokość)
        {
            pracownicy.Add(pracownik);
        }

        public IPracownik[] Pracownicy
        {
            get
            {
                return pracownicy.ToArray();
            }
        }

        public List<IPracownik> ListaPracowników
        {
            get
            {
                return pracownicy;
            }
        }
    }
}
